from typing import Final


__author__: Final = "Tsuyoshi Hombashi"
__copyright__: Final = f"Copyright 2020-2025, {__author__}"
__license__: Final = "MIT License"
__version__ = "0.6.3"
__maintainer__: Final = __author__
__email__: Final = "tsuyoshi.hombashi@gmail.com"



import difflib
import sys
from textwrap import dedent

import pytest
from pytablewriter.writer.text import MarkdownFlavor

from pytest_md_report.plugin import (
    ColorPolicy,
    extract_file_color_policy,
    is_apply_ansi_escape_to_file,
    is_apply_ansi_escape_to_term,
)


PYFILE_PASS_TEST = dedent(
    """\
    import pytest

    def test_pass():
        assert True
    """
)
PYFILE_SKIP_TEST = dedent(
    """\
    import pytest

    def test_skipped():
        pytest.skip()
    """
)
PYFILE_MIX_TESTS = dedent(
    """\
    import pytest

    def test_pass():
        assert True

    def test_failed():
        assert False

    def test_skipped():
        pytest.skip()

    def test_error(invalid_fixture):
        pass

    @pytest.mark.xfail()
    def test_xfailed():
        assert False

    @pytest.mark.xfail()
    def test_xpassed():
        assert True
    """
)


def print_test_result(expected, actual, error=None):
    print(f"[expected]\n{expected}\n")
    print(f"[actual]\n{actual}\n")

    if error:
        print(error, file=sys.stderr)

    print("----------------------------------------")
    d = difflib.Differ()
    diff = d.compare(expected.splitlines(), actual.splitlines())
    for d in diff:
        print(d)


@pytest.mark.parametrize(
    ["color_policy", "is_output_file", "expected"],
    [
        [ColorPolicy.AUTO, False, True],
        [ColorPolicy.AUTO, True, False],
        [ColorPolicy.TEXT, False, True],
        [ColorPolicy.TEXT, True, True],
        [ColorPolicy.NEVER, False, False],
        [ColorPolicy.NEVER, True, False],
    ],
)
def test_is_apply_ansi_escape_to_file(color_policy, is_output_file, expected):
    assert is_apply_ansi_escape_to_file(color_policy, is_output_file) == expected


@pytest.mark.parametrize(
    ["color_policy", "expected"],
    [
        [ColorPolicy.AUTO, True],
        [ColorPolicy.TEXT, True],
        [ColorPolicy.NEVER, False],
    ],
)
def test_is_apply_ansi_escape_to_term(color_policy, expected):
    assert is_apply_ansi_escape_to_term(color_policy) == expected


@pytest.mark.parametrize(
    ["color_policy", "is_output_file", "flavor", "expected"],
    [
        [ColorPolicy.AUTO, False, MarkdownFlavor.GFM, ColorPolicy.AUTO],
        [ColorPolicy.AUTO, True, MarkdownFlavor.GFM, ColorPolicy.AUTO],
        [ColorPolicy.AUTO, False, MarkdownFlavor.COMMON_MARK, ColorPolicy.AUTO],
        [ColorPolicy.AUTO, True, MarkdownFlavor.COMMON_MARK, ColorPolicy.NEVER],
        [ColorPolicy.TEXT, False, MarkdownFlavor.GFM, ColorPolicy.TEXT],
        [ColorPolicy.TEXT, True, MarkdownFlavor.GFM, ColorPolicy.TEXT],
        [ColorPolicy.NEVER, False, MarkdownFlavor.GFM, ColorPolicy.NEVER],
        [ColorPolicy.NEVER, True, MarkdownFlavor.GFM, ColorPolicy.NEVER],
    ],
)
def test_extract_file_color_policy(color_policy, is_output_file, flavor, expected):
    assert extract_file_color_policy(color_policy, is_output_file, flavor) == expected


def test_pytest_md_report(testdir):
    testdir.makepyfile(PYFILE_MIX_TESTS)
    expected = dedent(
        """\
        |         filepath         | passed | failed | error | skipped | xfailed | xpassed | SUBTOTAL |
        | ------------------------ | -----: | -----: | ----: | ------: | ------: | ------: | -------: |
        | test_pytest_md_report.py |      1 |      1 |     1 |       1 |       1 |       1 |        6 |
        | TOTAL                    |      1 |      1 |     1 |       1 |       1 |       1 |        6 |"""
    )
    result = testdir.runpytest("--md-report", "--md-report-color", "never")
    out = "\n".join(result.outlines[-4:])
    print_test_result(expected=expected, actual=out)

    assert out == expected


def test_pytest_md_report_output(testdir):
    testdir.makepyfile(PYFILE_MIX_TESTS)
    expected = dedent(
        """\
        |            filepath             | passed | failed | error | skipped | xfailed | xpassed | SUBTOTAL |
        | ------------------------------- | -----: | -----: | ----: | ------: | ------: | ------: | -------: |
        | test_pytest_md_report_output.py |      1 |      1 |     1 |       1 |       1 |       1 |        6 |
        | TOTAL                           |      1 |      1 |     1 |       1 |       1 |       1 |        6 |"""
    )
    output_filepath = testdir.tmpdir.join("report.md")
    result = testdir.runpytest(
        "--md-report",
        "--md-report-color",
        "never",
        "--md-report-output",
        output_filepath,
    )
    out = "\n".join(result.outlines[-4:])
    assert out != expected
    with open(output_filepath) as f:
        assert f.read().strip() == expected

    result = testdir.runpytest(
        "--md-report",
        "--md-report-color",
        "never",
        "--md-report-output",
        output_filepath,
        "--md-report-tee",
    )
    out = "\n".join(result.outlines[-4:])
    assert out == expected
    with open(output_filepath) as f:
        assert f.read().strip() == expected


def test_pytest_md_report_margin(testdir):
    testdir.makepyfile(PYFILE_MIX_TESTS)
    expected = dedent(
        """\
        |           filepath            |passed|failed|error|skipped|xfailed|xpassed|SUBTOTAL|
        |-------------------------------|-----:|-----:|----:|------:|------:|------:|-------:|
        |test_pytest_md_report_margin.py|     1|     1|    1|      1|      1|      1|       6|
        |TOTAL                          |     1|     1|    1|      1|      1|      1|       6|"""
    )
    result = testdir.runpytest(
        "--md-report", "--md-report-color", "never", "--md-report-margin", "0"
    )
    out = "\n".join(result.outlines[-4:])
    print_test_result(expected=expected, actual=out)

    assert out == expected


def test_pytest_md_report_zeros(testdir):
    testdir.makepyfile(test_passed=PYFILE_PASS_TEST)
    testdir.makepyfile(test_skipped=PYFILE_SKIP_TEST)

    expected = dedent(
        """\
        |    filepath     | passed | skipped | SUBTOTAL |
        | --------------- | -----: | ------: | -------: |
        | test_passed.py  |      1 |         |        1 |
        | test_skipped.py |        |       1 |        1 |
        | TOTAL           |      1 |       1 |        2 |"""
    )
    result = testdir.runpytest(
        "--md-report", "--md-report-color", "never", "--md-report-zeros", "empty"
    )
    out = "\n".join(result.outlines[-5:])
    print_test_result(expected=expected, actual=out)

    assert out == expected


@pytest.mark.parametrize(
    ["color_option"],
    [
        ["--md-report-success-color"],
        ["--md-report-skip-color"],
        ["--md-report-error-color"],
    ],
)
def test_pytest_md_report_results_color(testdir, color_option):
    testdir.makepyfile(PYFILE_MIX_TESTS)
    org_out = "\n".join(testdir.runpytest("--md-report").outlines[-4:])
    ch_color_out = "\n".join(
        testdir.runpytest("--md-report", color_option, "#ff2a2a").outlines[-4:]
    )

    assert org_out != ch_color_out


def test_pytest_md_report_flavor(testdir):
    testdir.makepyfile(test_passed=PYFILE_PASS_TEST)
    testdir.makepyfile(test_skipped=PYFILE_SKIP_TEST)
    out_dir = testdir.mkdir("outputs")
    out_file = out_dir.join("report.md")
    testdir.runpytest("--md-report", "--md-report-flavor", "github", "--md-report-output", out_file)

    with open(out_file) as f:
        report = f.read()
        print(report)
        assert (
            report
            == r"""|    filepath     | $$\textcolor{#23d18b}{\tt{passed}}$$ | $$\textcolor{#f5f543}{\tt{skipped}}$$ | SUBTOTAL |
| --------------- | --------------------------------: | --------------------------------: | -------: |
| $$\textcolor{#23d18b}{\tt{test\\_passed.py}}$$ |   $$\textcolor{#23d18b}{\tt{1}}$$ |   $$\textcolor{#666666}{\tt{0}}$$ | $$\textcolor{#23d18b}{\tt{1}}$$ |
| $$\textcolor{#f5f543}{\tt{test\\_skipped.py}}$$ |   $$\textcolor{#666666}{\tt{0}}$$ |   $$\textcolor{#f5f543}{\tt{1}}$$ | $$\textcolor{#f5f543}{\tt{1}}$$ |
| $$\textcolor{#f5f543}{\tt{TOTAL}}$$ |   $$\textcolor{#23d18b}{\tt{1}}$$ |   $$\textcolor{#f5f543}{\tt{1}}$$ | $$\textcolor{#f5f543}{\tt{2}}$$ |
"""
        )


def test_pytest_md_report_exclude_outcomes(testdir):
    testdir.makepyfile(PYFILE_MIX_TESTS)
    expected = dedent(
        """\
        |                 filepath                  | failed | error | xfailed | SUBTOTAL |
        | ----------------------------------------- | -----: | ----: | ------: | -------: |
        | test_pytest_md_report_exclude_outcomes.py |      1 |     1 |       1 |        3 |
        | TOTAL                                     |      1 |     1 |       1 |        6 |
        """
    )
    output_filepath = testdir.tmpdir.join("report.md")
    testdir.runpytest(
        "--md-report",
        "--md-report-exclude-outcomes",
        "passed",
        "skipped",
        "xpassed",
        "--md-report-output",
        output_filepath,
    )
    with open(output_filepath) as f:
        report = f.read()
        print(report)
        assert report == expected



from .__version__ import __author__, __copyright__, __email__, __license__, __version__
from ._const import ColorPolicy, ZerosRender
from .plugin import make_md_report, retrieve_stat_count_map


__all__ = (
    "ColorPolicy",
    "ZerosRender",
    "make_md_report",
    "retrieve_stat_count_map",
    "__author__",
    "__copyright__",
    "__email__",
    "__license__",
    "__version__",
)



import os.path
import re
from typing import Final

import setuptools


MODULE_NAME: Final = "pytest-md-report"
REPOSITORY_URL: Final = f"https://github.com/thombashi/{MODULE_NAME:s}"
REQUIREMENT_DIR: Final = "requirements"
ENCODING: Final = "utf8"

pkg_info: dict[str, str] = {}


def get_release_command_class() -> dict[str, type[setuptools.Command]]:
    try:
        from releasecmd import ReleaseCommand
    except ImportError:
        return {}

    return {"release": ReleaseCommand}


def make_long_description() -> str:
    # ref: https://github.com/pypa/readme_renderer/issues/304
    re_exclude = re.compile(r"\s*:scale:\s*\d+")

    with open("README.rst", encoding=ENCODING) as f:
        return "".join([line for line in f if not re_exclude.search(line)])


with open(os.path.join(MODULE_NAME.replace("-", "_"), "__version__.py")) as f:
    exec(f.read(), pkg_info)

with open(os.path.join(REQUIREMENT_DIR, "requirements.txt")) as f:
    INSTALL_REQUIRES = [line.strip() for line in f if line.strip()]

with open(os.path.join(REQUIREMENT_DIR, "test_requirements.txt")) as f:
    TESTS_REQUIRES = [line.strip() for line in f if line.strip()]


setuptools.setup(
    name=MODULE_NAME,
    url=REPOSITORY_URL,
    author=pkg_info["__author__"],
    author_email=pkg_info["__email__"],
    description="A pytest plugin to generate test outcomes reports with markdown table format.",
    include_package_data=True,
    keywords=["pytest", "plugin", "markdown"],
    license=pkg_info["__license__"],
    long_description=make_long_description(),
    long_description_content_type="text/x-rst",
    packages=setuptools.find_packages(),
    package_data={MODULE_NAME: ["py.typed"]},
    project_urls={
        "Changelog": f"{REPOSITORY_URL:s}/releases",
        "Source": REPOSITORY_URL,
        "Tracker": f"{REPOSITORY_URL:s}/issues",
    },
    python_requires=">=3.9",
    install_requires=INSTALL_REQUIRES,
    extras_require={"test": TESTS_REQUIRES},
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Plugins",
        "Framework :: Pytest",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Software Development :: Quality Assurance",
        "Topic :: Software Development :: Testing",
        "Topic :: Text Processing",
        "Topic :: Text Processing :: Markup :: Markdown",
        "Typing :: Typed",
    ],
    cmdclass=get_release_command_class(),
    zip_safe=False,
    entry_points={
        "pytest11": [
            "pytest-md-report = pytest_md_report.plugin",
        ]
    },
)



from enum import Enum, unique
from textwrap import dedent
from typing import Final

from pytablewriter.writer.text import MarkdownFlavor
from tcolorpy import AnsiFGColor


COLOR_NAMES: Final = "/".join([style.name.lower() for style in list(AnsiFGColor)])


class Header:
    FILEPATH: Final = "filepath"
    TESTFUNC: Final = "function"
    SUBTOTAL: Final = "SUBTOTAL"


class ColorPolicy(Enum):
    AUTO = "auto"
    TEXT = "text"
    NEVER = "never"


class ZerosRender:
    NUMBER: Final = "number"
    EMPTY: Final = "empty"
    LIST: Final = (NUMBER, EMPTY)


class FGColor:
    SUCCESS: Final = "SUCCESS"
    ERROR: Final = "ERROR"
    SKIP: Final = "SKIP"
    GRAYOUT: Final = "GRAYOUT"


class BGColor:
    EVEN_ROW: Final = "#202020"
    ODD_ROW: Final = "black"
    TOTAL_ROW: Final = "#000000"


class Default:
    COLOR_POLICY: Final = ColorPolicy.AUTO
    MARGIN: Final = 1
    MARKDOWN_FLAVOR: Final = MarkdownFlavor.COMMON_MARK
    ZEROS: Final = ZerosRender.NUMBER
    EXCLUDE_RESULTS: list[str] = []

    class FGColor:
        SUCCESS: Final = "light_green"
        ERROR: Final = "light_red"
        SKIP: Final = "light_yellow"
        GRAYOUT: Final = "light_black"


OPTION_PREFIX: Final = "md-report"


@unique
class Option(Enum):
    MD_REPORT = (OPTION_PREFIX, "Create a Markdown report.")
    MD_REPORT_VERBOSE = (
        f"{OPTION_PREFIX}-verbose",
        dedent(
            """\
            Verbosity level for pytest-md-report.
            If not set, use the verbosity level of pytest.
            Defaults to 0.
            """
        ),
    )
    MD_REPORT_OUTPUT = (
        f"{OPTION_PREFIX}-output",
        dedent(
            """\
            Path to a file to the outputs test report.
            Overwrite a file content if the file already exists.
            """
        ),
    )
    MD_REPORT_TEE = (
        f"{OPTION_PREFIX}-tee",
        "output test report for both standard output and a file.",
    )
    MD_REPORT_COLOR = (
        f"{OPTION_PREFIX}-color",
        dedent(
            """\
            How coloring output reports.
            auto: detect the output destination and colorize reports appropriately with the output.
            for terminal output, render colored (text and background) reports using ANSI escape codes.
            for file output, render the report without color.
            text: render colored text reports by using ANSI escape codes.
            never: render report without color.
            Defaults to '{default}'.
            """
        ).format(default=Default.COLOR_POLICY.value),
    )
    MD_REPORT_MARGIN = (
        f"{OPTION_PREFIX}-margin",
        dedent(
            """\
            Margin size for each cell.
            Defaults to {default}.
            """
        ).format(default=Default.MARGIN),
    )
    MD_REPORT_ZEROS = (
        f"{OPTION_PREFIX}-zeros",
        dedent(
            """\
            Rendering method for results of zero values.
            number: render as a digit number (0).
            empty: not rendering.
            Automatically set to 'number' when the CI environment variable is set to
            TRUE (case insensitive) to display reports correctly at CI services.
            Defaults to '{default}'.
            """
        ).format(default=Default.ZEROS),
    )
    MD_REPORT_SUCCESS_COLOR = (
        f"{OPTION_PREFIX}-success-color",
        dedent(
            """\
            Text color of succeeded results.
            Specify a color name (one of the {names}) or a color code (e.g. #ff1020).
            Defaults to '{default}'.
            """
        ).format(names=COLOR_NAMES, default=Default.FGColor.SUCCESS),
    )
    MD_REPORT_SKIP_COLOR = (
        f"{OPTION_PREFIX}-skip-color",
        dedent(
            """\
            Text color of skipped results.
            Specify a color name (one of the {names}) or a color code (e.g. #ff1020).
            Defaults to '{default}'.
            """
        ).format(names=COLOR_NAMES, default=Default.FGColor.SKIP),
    )
    MD_REPORT_ERROR_COLOR = (
        f"{OPTION_PREFIX}-error-color",
        dedent(
            """\
            Text color of failed results.
            Specify a color name (one of the {names}) or a color code (e.g. #ff1020).
            Defaults to '{default}'.
            """
        ).format(names=COLOR_NAMES, default=Default.FGColor.ERROR),
    )
    MD_REPORT_FLAVOR = (
        f"{OPTION_PREFIX}-flavor",
        dedent(
            """\
            Markdown flavor of the output report.
            Defaults to '{default}'.
            """
        ).format(
            default=Default.MARKDOWN_FLAVOR.value,
        ),
    )
    MD_EXCLUDE_OUTCOMES = (
        f"{OPTION_PREFIX}-exclude-outcomes",
        dedent(
            """\
            List of test outcomes to exclude from the report.
            When specifying as an environment variable, pass a comma-separated string
            (e.g. 'passed,skipped').
            Defaults to '{default}'.
            """
        ).format(
            default=Default.EXCLUDE_RESULTS,
        ),
    )

    @property
    def cmdoption_str(self) -> str:
        return "--" + self.__name.lower()

    @property
    def envvar_str(self) -> str:
        return "PYTEST_" + self.__name.replace("-", "_").upper()

    @property
    def inioption_str(self) -> str:
        return self.__name.replace("-", "_").lower()

    @property
    def help_msg(self) -> str:
        return self.__help_msg

    def __init__(self, name: str, help_msg: str) -> None:
        self.__name = name.strip()
        self.__help_msg = help_msg


class HelpMsg:
    EXTRA_MSG_TEMPLATE: Final = " you can also specify the value with {} environment variable."



import os
from collections import defaultdict
from collections.abc import Mapping, Sequence
from typing import Any, Optional

from _pytest.config import Config
from _pytest.config.argparsing import Parser
from _pytest.terminal import TerminalReporter
from pytablewriter import TableWriterFactory
from pytablewriter.writer.text import MarkdownFlavor, normalize_md_flavor
from typepy import Bool, Integer, StrictLevel, is_not_null_string
from typepy.error import TypeConversionError

from ._const import ColorPolicy, Default, FGColor, Header, HelpMsg, Option, ZerosRender
from ._style_filter import col_separator_style_filter, style_filter


def zero_to_nullstr(value: Any) -> Any:
    if value == 0:
        return ""

    return value


def pytest_addoption(parser: Parser) -> None:
    group = parser.getgroup("md report", "generate test outcomes report with markdown table format")

    group.addoption(
        Option.MD_REPORT.cmdoption_str,
        action="store_true",
        default=None,
        help=Option.MD_REPORT.help_msg
        + HelpMsg.EXTRA_MSG_TEMPLATE.format(Option.MD_REPORT.envvar_str),
    )
    group.addoption(
        Option.MD_REPORT_VERBOSE.cmdoption_str,
        metavar="VERBOSITY_LEVEL",
        type=int,
        default=None,
        help=Option.MD_REPORT_VERBOSE.help_msg
        + HelpMsg.EXTRA_MSG_TEMPLATE.format(Option.MD_REPORT_VERBOSE.envvar_str),
    )
    group.addoption(
        Option.MD_REPORT_OUTPUT.cmdoption_str,
        metavar="FILEPATH",
        default=None,
        help=Option.MD_REPORT_OUTPUT.help_msg
        + HelpMsg.EXTRA_MSG_TEMPLATE.format(Option.MD_REPORT_OUTPUT.envvar_str),
    )
    group.addoption(
        Option.MD_REPORT_TEE.cmdoption_str,
        action="store_true",
        default=None,
        help=Option.MD_REPORT_TEE.help_msg
        + HelpMsg.EXTRA_MSG_TEMPLATE.format(Option.MD_REPORT_TEE.envvar_str),
    )
    group.addoption(
        Option.MD_REPORT_COLOR.cmdoption_str,
        choices=[policy.value for policy in ColorPolicy],
        default=None,
        help=Option.MD_REPORT_COLOR.help_msg
        + HelpMsg.EXTRA_MSG_TEMPLATE.format(Option.MD_REPORT_COLOR.envvar_str),
    )
    group.addoption(
        Option.MD_REPORT_MARGIN.cmdoption_str,
        metavar="MARGIN",
        type=int,
        default=None,
        help=Option.MD_REPORT_MARGIN.help_msg
        + HelpMsg.EXTRA_MSG_TEMPLATE.format(Option.MD_REPORT_MARGIN.envvar_str),
    )
    group.addoption(
        Option.MD_REPORT_ZEROS.cmdoption_str,
        choices=ZerosRender.LIST,
        default=None,
        help=Option.MD_REPORT_ZEROS.help_msg
        + HelpMsg.EXTRA_MSG_TEMPLATE.format(Option.MD_REPORT_ZEROS.envvar_str),
    )
    group.addoption(
        Option.MD_REPORT_SUCCESS_COLOR.cmdoption_str,
        default=None,
        help=Option.MD_REPORT_SUCCESS_COLOR.help_msg
        + HelpMsg.EXTRA_MSG_TEMPLATE.format(Option.MD_REPORT_SUCCESS_COLOR.envvar_str),
    )
    group.addoption(
        Option.MD_REPORT_SKIP_COLOR.cmdoption_str,
        default=None,
        help=Option.MD_REPORT_SKIP_COLOR.help_msg
        + HelpMsg.EXTRA_MSG_TEMPLATE.format(Option.MD_REPORT_SKIP_COLOR.envvar_str),
    )
    group.addoption(
        Option.MD_REPORT_ERROR_COLOR.cmdoption_str,
        default=None,
        help=Option.MD_REPORT_ERROR_COLOR.help_msg
        + HelpMsg.EXTRA_MSG_TEMPLATE.format(Option.MD_REPORT_ERROR_COLOR.envvar_str),
    )
    group.addoption(
        Option.MD_REPORT_FLAVOR.cmdoption_str,
        choices=[flavor.value for flavor in MarkdownFlavor],
        default=None,
        help=Option.MD_REPORT_FLAVOR.help_msg
        + HelpMsg.EXTRA_MSG_TEMPLATE.format(Option.MD_REPORT_FLAVOR.envvar_str),
    )
    group.addoption(
        Option.MD_EXCLUDE_OUTCOMES.cmdoption_str,
        nargs="+",
        default=[],
        help=Option.MD_EXCLUDE_OUTCOMES.help_msg
        + HelpMsg.EXTRA_MSG_TEMPLATE.format(Option.MD_EXCLUDE_OUTCOMES.envvar_str),
    )

    parser.addini(
        Option.MD_REPORT.inioption_str,
        type="bool",
        default=False,
        help=Option.MD_REPORT.help_msg,
    )
    parser.addini(
        Option.MD_REPORT_VERBOSE.inioption_str,
        default=None,
        help=Option.MD_REPORT_VERBOSE.help_msg,
    )
    parser.addini(
        Option.MD_REPORT_COLOR.inioption_str,
        default=None,
        help=Option.MD_REPORT_COLOR.help_msg,
    )
    parser.addini(
        Option.MD_REPORT_OUTPUT.inioption_str,
        default=None,
        help=Option.MD_REPORT_OUTPUT.help_msg,
    )
    parser.addini(
        Option.MD_REPORT_TEE.inioption_str,
        default=None,
        help=Option.MD_REPORT_TEE.help_msg,
    )
    parser.addini(
        Option.MD_REPORT_MARGIN.inioption_str,
        default=None,
        help=Option.MD_REPORT_MARGIN.help_msg,
    )
    parser.addini(
        Option.MD_REPORT_ZEROS.inioption_str,
        default=None,
        help=Option.MD_REPORT_ZEROS.help_msg,
    )
    parser.addini(
        Option.MD_REPORT_SUCCESS_COLOR.inioption_str,
        default=None,
        help=Option.MD_REPORT_SUCCESS_COLOR.help_msg,
    )
    parser.addini(
        Option.MD_REPORT_SKIP_COLOR.inioption_str,
        default=None,
        help=Option.MD_REPORT_SKIP_COLOR.help_msg,
    )
    parser.addini(
        Option.MD_REPORT_ERROR_COLOR.inioption_str,
        default=None,
        help=Option.MD_REPORT_ERROR_COLOR.help_msg,
    )
    parser.addini(
        Option.MD_REPORT_FLAVOR.inioption_str,
        default=None,
        help=Option.MD_REPORT_FLAVOR.help_msg,
    )
    parser.addini(
        Option.MD_EXCLUDE_OUTCOMES.inioption_str,
        type="args",
        default=[],
        help=Option.MD_EXCLUDE_OUTCOMES.help_msg,
    )


def is_make_md_report(config: Config) -> bool:
    if config.option.help:
        return False

    make_report: Optional[bool] = config.option.md_report

    if make_report is None:
        make_report = _to_bool(os.environ.get(Option.MD_REPORT.envvar_str))

    if make_report is None:
        make_report = _to_bool(config.getini(Option.MD_REPORT.inioption_str))

    if make_report is None:
        return False

    return make_report


def is_apply_ansi_escape_to_file(color_policy: ColorPolicy, is_output_file: bool) -> bool:
    if color_policy == ColorPolicy.TEXT:
        return True

    if color_policy == ColorPolicy.AUTO and not is_output_file:
        return True

    return False


def is_apply_ansi_escape_to_term(color_policy: ColorPolicy) -> bool:
    return color_policy in [ColorPolicy.TEXT, ColorPolicy.AUTO]


def _is_ci() -> bool:
    # many CI services will set the CI environment variable to 'true'
    CI = os.environ.get("CI")
    if not CI:
        return False

    return CI.lower() == "true"


def _is_travis_ci() -> bool:
    # https://docs.travis-ci.com/user/environment-variables/#default-environment-variables
    return os.environ.get("TRAVIS") == "true"


def _is_appveyor_ci() -> bool:
    # https://www.appveyor.com/docs/environment-variables/
    APPVEYOR = os.environ.get("APPVEYOR")
    if not APPVEYOR:
        return False

    return APPVEYOR.lower() == "true"


def _to_int(value: Any) -> Optional[int]:
    try:
        return Integer(value, strict_level=StrictLevel.MIN).convert()
    except TypeConversionError:
        pass

    return None


def _to_bool(value: Any) -> Optional[bool]:
    try:
        return Bool(value, strict_level=StrictLevel.MIN).convert()
    except TypeConversionError:
        pass

    return None


def retrieve_verbosity_level(config: Config) -> int:
    verbosity_level: Optional[int] = config.option.md_report_verbose

    if verbosity_level is not None and verbosity_level < 0:
        verbosity_level = None

    if verbosity_level is None:
        verbosity_level = _to_int(os.environ.get(Option.MD_REPORT_VERBOSE.envvar_str))

    if verbosity_level is None:
        verbosity_level = _to_int(config.getini(Option.MD_REPORT_VERBOSE.inioption_str))

    if verbosity_level is None:
        # use the verbosity level of the pytest if not set
        return config.option.verbose

    return verbosity_level


def retrieve_output_filepath(config: Config) -> Optional[str]:
    output_filepath: Optional[str] = config.option.md_report_output

    if not output_filepath:
        output_filepath = os.environ.get(Option.MD_REPORT_OUTPUT.envvar_str)

    if not output_filepath:
        value = config.getini(Option.MD_REPORT_OUTPUT.inioption_str)
        if value is None:
            return None
        else:
            return str(value)

    return output_filepath


def retrieve_tee(config: Config) -> bool:
    tee: Optional[bool] = config.option.md_report_tee

    if tee is None:
        envvar_str = os.environ.get(Option.MD_REPORT_TEE.envvar_str)
        if envvar_str is not None:
            tee = bool(envvar_str)

    if tee is None:
        inioption_str = config.getini(Option.MD_REPORT_TEE.inioption_str)
        if inioption_str is not None:
            tee = bool(inioption_str)

    return tee if tee is not None else False


def retrieve_md_flavor(config: Config) -> MarkdownFlavor:
    md_flavor = config.option.md_report_flavor

    if not md_flavor:
        md_flavor = os.environ.get(Option.MD_REPORT_FLAVOR.envvar_str)

    if not md_flavor:
        md_flavor = config.getini(Option.MD_REPORT_FLAVOR.inioption_str)

    if not md_flavor:
        return Default.MARKDOWN_FLAVOR

    return normalize_md_flavor(str(md_flavor))


def retrieve_exclude_outcomes(config: Config) -> list[str]:
    def norm_names(names: Sequence[Any]) -> list[str]:
        return [str(name).lower().strip() for name in names]

    exclude_outcomes = config.option.md_report_exclude_outcomes

    if not exclude_outcomes:
        exclude_outcomes = os.environ.get(Option.MD_EXCLUDE_OUTCOMES.envvar_str)
        if exclude_outcomes:
            return norm_names(exclude_outcomes.split(","))

    if not exclude_outcomes:
        exclude_outcomes = config.getini(Option.MD_EXCLUDE_OUTCOMES.inioption_str)

    if not exclude_outcomes:
        return Default.EXCLUDE_RESULTS

    if isinstance(exclude_outcomes, list):
        # list will be passed via pytest config file
        return norm_names(exclude_outcomes)
    elif isinstance(exclude_outcomes, str):
        # comma-separated string (e.g. passed,skipped) will be passed via the command line option
        return norm_names(exclude_outcomes.split(","))

    raise TypeError(f"Unexpected type {type(exclude_outcomes)} for exclude_outcomes")


def retrieve_color_policy(config: Config) -> ColorPolicy:
    color_policy = config.option.md_report_color

    if not color_policy:
        color_policy = os.environ.get(Option.MD_REPORT_COLOR.envvar_str)

    if not color_policy:
        color_policy = config.getini(Option.MD_REPORT_COLOR.inioption_str)

    if not color_policy:
        return Default.COLOR_POLICY

    return ColorPolicy[str(color_policy).upper()]


def retrieve_report_margin(config: Config) -> int:
    margin = config.option.md_report_margin

    if margin is None:
        margin = _to_int(os.environ.get(Option.MD_REPORT_MARGIN.envvar_str))

    if margin is None:
        margin = _to_int(config.getini(Option.MD_REPORT_MARGIN.inioption_str))

    if margin is None:
        return Default.MARGIN

    return margin


def retrieve_report_zeros(config: Config) -> str:
    report_zeros = config.option.md_report_zeros

    if not report_zeros:
        report_zeros = os.environ.get(Option.MD_REPORT_ZEROS.envvar_str)

    if not report_zeros:
        report_zeros = config.getini(Option.MD_REPORT_ZEROS.inioption_str)

    if _is_ci():
        report_zeros = ZerosRender.NUMBER

    if not report_zeros:
        report_zeros = Default.ZEROS

    return str(report_zeros)


def retrieve_report_results_color(config: Config, color_option: Option, default: str) -> str:
    results_color = getattr(config.option, color_option.inioption_str)

    if not results_color:
        results_color = os.environ.get(color_option.envvar_str)

    if not results_color:
        results_color = config.getini(color_option.inioption_str)

    if not results_color:
        results_color = default

    return str(results_color)


def _normalize_stat_name(name: str) -> str:
    if name == "error":
        return "errors"

    return name


def retrieve_stat_count_map(reporter: TerminalReporter) -> dict[str, int]:
    stat_count_map = {}

    for name in ["failed", "passed", "skipped", "error", "xfailed", "xpassed"]:
        count = len(reporter.getreports(name))
        stat_count_map[name] = count

    return stat_count_map


def extract_pytest_stats(
    reporter: TerminalReporter, outcomes: Sequence[str], verbosity_level: int
) -> Mapping[tuple, Mapping[str, int]]:
    results_per_testfunc: dict[tuple, dict[str, int]] = {}

    for stat_key, values in reporter.stats.items():
        if stat_key not in outcomes:
            continue

        for value in values:
            try:
                filesystempath, lineno, domaininfo = value.location
            except AttributeError:
                continue

            filesystempath = os.path.normpath(filesystempath).replace("\\", "/")
            testfunc = value.head_line.split("[")[0]

            if verbosity_level == 0:
                key: tuple = (filesystempath,)
            elif verbosity_level >= 1:
                key = (filesystempath, testfunc)
            else:
                continue

            if key not in results_per_testfunc:
                results_per_testfunc[key] = defaultdict(int)

            results_per_testfunc[key][stat_key] += 1

    return results_per_testfunc


def make_md_report(
    config: Config,
    reporter: TerminalReporter,
    total_stats: Mapping[str, int],
    color_policy: ColorPolicy,
    apply_ansi_escape: bool,
    md_flavor: MarkdownFlavor,
) -> str:
    verbosity_level = retrieve_verbosity_level(config)
    exclude_outcomes = retrieve_exclude_outcomes(config)

    outcomes = ["passed", "failed", "error", "skipped", "xfailed", "xpassed"]
    outcomes = [key for key in outcomes if key not in exclude_outcomes]
    outcomes = [key for key in outcomes if total_stats.get(key, 0) > 0]

    if not outcomes:
        return ""

    results_per_testfunc = extract_pytest_stats(
        reporter=reporter, outcomes=outcomes, verbosity_level=verbosity_level
    )

    writer = TableWriterFactory.create_from_format_name(
        "md", flavor=md_flavor.value, colorize_terminal=apply_ansi_escape
    )

    matrix = [
        list(key) + [results.get(key, 0) for key in outcomes] + [sum(results.values())]
        for key, results in results_per_testfunc.items()
    ]
    if verbosity_level == 0:
        writer.headers = [Header.FILEPATH] + outcomes + [Header.SUBTOTAL]
        matrix.append(
            ["TOTAL"] + [total_stats.get(key, 0) for key in outcomes] + [sum(total_stats.values())]
        )
    elif verbosity_level >= 1:
        writer.headers = [Header.FILEPATH, Header.TESTFUNC] + outcomes + [Header.SUBTOTAL]
        matrix.append(
            ["TOTAL", ""]
            + [total_stats.get(key, 0) for key in outcomes]
            + [sum(total_stats.values())]
        )

    writer.margin = retrieve_report_margin(config)
    writer.value_matrix = matrix

    if color_policy != ColorPolicy.NEVER:
        writer.style_filter_kwargs = {
            "color_policy": color_policy,
            "color_map": {
                FGColor.SUCCESS: retrieve_report_results_color(
                    config, Option.MD_REPORT_SUCCESS_COLOR, Default.FGColor.SUCCESS
                ),
                FGColor.ERROR: retrieve_report_results_color(
                    config, Option.MD_REPORT_ERROR_COLOR, Default.FGColor.ERROR
                ),
                FGColor.SKIP: retrieve_report_results_color(
                    config, Option.MD_REPORT_SKIP_COLOR, Default.FGColor.SKIP
                ),
                FGColor.GRAYOUT: Default.FGColor.GRAYOUT,
            },
            "num_rows": len(writer.value_matrix),
        }

        if not _is_travis_ci():
            writer.add_style_filter(style_filter)

        if color_policy == ColorPolicy.AUTO and not _is_ci():
            writer.add_col_separator_style_filter(col_separator_style_filter)

    report_zeros = retrieve_report_zeros(config)
    if report_zeros == ZerosRender.EMPTY:
        writer.register_trans_func(zero_to_nullstr)

    return writer.dumps()


def extract_file_color_policy(
    color_policy: ColorPolicy, is_output_file: bool, md_flavor: MarkdownFlavor
) -> ColorPolicy:
    file_color_policy = color_policy
    if is_output_file and color_policy == ColorPolicy.AUTO:
        if md_flavor != MarkdownFlavor.GFM:
            file_color_policy = ColorPolicy.NEVER

    return file_color_policy


def pytest_unconfigure(config: Config) -> None:
    if not is_make_md_report(config):
        return

    reporter = config.pluginmanager.get_plugin("terminalreporter")
    if reporter is None:
        return

    stat_count_map = retrieve_stat_count_map(reporter)
    is_tee = retrieve_tee(config)
    output_filepath = retrieve_output_filepath(config)
    color_policy = retrieve_color_policy(config)
    md_flavor = retrieve_md_flavor(config)

    is_output_term = is_tee or not output_filepath
    is_output_file = is_not_null_string(output_filepath)
    term_color_policy = color_policy
    file_color_policy = extract_file_color_policy(color_policy, is_output_file, md_flavor)

    apply_ansi_escape_to_file = is_apply_ansi_escape_to_file(color_policy, is_output_file)
    apply_ansi_escape_to_term = is_apply_ansi_escape_to_term(color_policy)
    term_report = None
    if is_output_term:
        term_report = make_md_report(
            config,
            reporter,
            stat_count_map,
            color_policy=term_color_policy,
            apply_ansi_escape=apply_ansi_escape_to_term,
            md_flavor=md_flavor,
        )
        reporter._tw.write(term_report)

    if not is_output_file:
        return

    file_report = term_report
    gen_report_for_file_output = any(
        [
            not file_report,
            file_color_policy != term_color_policy,
            apply_ansi_escape_to_file != apply_ansi_escape_to_term,
        ]
    )
    if gen_report_for_file_output:
        file_report = make_md_report(
            config,
            reporter,
            stat_count_map,
            color_policy=file_color_policy,
            apply_ansi_escape=apply_ansi_escape_to_file,
            md_flavor=md_flavor,
        )

    if file_report:
        assert output_filepath
        with open(output_filepath, "w") as f:
            f.write(file_report)



from typing import Any, Final, Optional, cast

from pytablewriter.style import Cell, Style
from pytablewriter.writer import AbstractTableWriter

from ._const import BGColor, ColorPolicy, FGColor, Header


class ColorRetriever:
    def __init__(
        self,
        row: int,
        is_grayout: bool,
        color_polilcy: ColorPolicy,
        color_map: dict[str, str],
    ) -> None:
        self.__row = row
        self.__is_grayout = is_grayout
        self.__color_polilcy = color_polilcy
        self.__color_map = color_map

    def retrieve_fg_bg_color(self, base_color: str) -> tuple[str, Optional[str]]:
        bg_color: Optional[str] = None

        if (self.__row % 2) == 0:
            fg_color = self.__color_map[FGColor.GRAYOUT] if self.__is_grayout else base_color
            bg_color = BGColor.EVEN_ROW if self.__color_polilcy == ColorPolicy.AUTO else None
        else:
            fg_color = self.__color_map[FGColor.GRAYOUT] if self.__is_grayout else base_color
            bg_color = BGColor.ODD_ROW if self.__color_polilcy == ColorPolicy.AUTO else None

        return (fg_color, bg_color)


def style_filter(cell: Cell, **kwargs: Any) -> Optional[Style]:
    writer = cast(AbstractTableWriter, kwargs["writer"])
    color_policy: Final = cast(ColorPolicy, kwargs["color_policy"])
    color_map: Final = kwargs["color_map"]
    num_rows: Final = cast(int, kwargs["num_rows"])
    headers: Final = writer.headers
    header: Final = headers[cell.col]
    fg_color = None
    bg_color = None

    is_grayout = False
    if cell.value == 0:
        is_grayout = True

    retriever = ColorRetriever(cell.row, is_grayout, color_policy, color_map)

    if cell.is_header_row():
        if all([writer.value_matrix[r][cell.col] == 0 for r in range(len(writer.value_matrix))]):
            return Style(color=color_map[FGColor.GRAYOUT])

        if header in ("passed"):
            fg_color, bg_color = retriever.retrieve_fg_bg_color(color_map[FGColor.SUCCESS])
        elif header in ("failed", "error"):
            fg_color, bg_color = retriever.retrieve_fg_bg_color(color_map[FGColor.ERROR])
        elif header in ("skipped", "xfailed", "xpassed"):
            fg_color, bg_color = retriever.retrieve_fg_bg_color(color_map[FGColor.SKIP])
        else:
            return None

        return Style(color=fg_color)

    is_passed = False
    is_failed = False
    is_skipped = False

    if header in (Header.FILEPATH, Header.TESTFUNC, Header.SUBTOTAL):
        error_ct_list = []
        if "failed" in headers:
            error_ct_list.append(writer.value_matrix[cell.row][headers.index("failed")])
        if "error" in headers:
            error_ct_list.append(writer.value_matrix[cell.row][headers.index("error")])

        skip_ct_list = []
        if "skipped" in headers:
            skip_ct_list.append(writer.value_matrix[cell.row][headers.index("skipped")])
        if "xfailed" in headers:
            skip_ct_list.append(writer.value_matrix[cell.row][headers.index("xfailed")])
        if "xpassed" in headers:
            skip_ct_list.append(writer.value_matrix[cell.row][headers.index("xpassed")])

        error_ct = sum(error_ct_list)
        skip_ct = sum(skip_ct_list)

        is_failed = error_ct > 0
        is_skipped = skip_ct > 0
        is_passed = error_ct == 0 and skip_ct == 0

    if is_passed or header in ("passed"):
        fg_color, bg_color = retriever.retrieve_fg_bg_color(color_map[FGColor.SUCCESS])
    elif is_failed or header in ("failed", "error"):
        fg_color, bg_color = retriever.retrieve_fg_bg_color(color_map[FGColor.ERROR])
    elif is_skipped or header in ("skipped", "xfailed", "xpassed"):
        fg_color, bg_color = retriever.retrieve_fg_bg_color(color_map[FGColor.SKIP])

    if cell.row == num_rows - 1:
        bg_color = BGColor.TOTAL_ROW

    return Style(color=fg_color, bg_color=bg_color)


def col_separator_style_filter(
    left_cell: Optional[Cell], right_cell: Optional[Cell], **kwargs: dict[str, Any]
) -> Optional[Style]:
    num_rows: Final = cast(int, kwargs["num_rows"])
    fg_color = None
    bg_color = None
    row: Final = left_cell.row if left_cell else cast(Cell, right_cell).row

    if row == num_rows - 1:
        bg_color = BGColor.TOTAL_ROW
    elif row % 2 == 0:
        bg_color = BGColor.EVEN_ROW
    elif row >= 0:
        bg_color = BGColor.ODD_ROW

    if fg_color or bg_color:
        return Style(color=fg_color, bg_color=bg_color)

    return None
